@@ -40,9 +40,12 @@ import sys
 import time
 import binascii
 
+# Set to False to not use async functions even though aiodns is installed.
+USE_ASYNC = True
+
 # only needed for arc
 try:
-    from authres import AuthenticationResultsHeader
+    import authres
 except ImportError:
     pass
 
@@ -72,13 +75,17 @@ from dkim.crypto import (
 try:
     from dkim.dnsplug import get_txt
 except ImportError:
-    try:
-        import aiodns
-        from dkim.asyncsupport import get_txt_async as get_txt
-    except:
-        # Only true if not using async
-        def get_txt(s,timeout=5):
-            raise RuntimeError("DKIM.verify requires DNS or dnspython module")
+    if USE_ASYNC:
+        try:
+            import aiodns
+            from dkim.asyncsupport import get_txt_async as get_txt
+        except:
+            # Only true if not using async
+            def get_txt(s,timeout=5):
+                raise RuntimeError("DKIM.verify requires DNS or dnspython module")
+    else:
+        raise RuntimeError("DKIM.verify requires DNS or dnspython module")
+
 from dkim.util import (
     get_default_logger,
     InvalidTagValueList,
@@ -94,6 +101,8 @@ __all__ = [
     "ValidationError",
     "AuthresNotFoundError",
     "NaClNotFoundError",
+    "DnsTimeoutError",
+    "USE_ASYNC",
     "CV_Pass",
     "CV_Fail",
     "CV_None",
@@ -187,6 +196,9 @@ class NaClNotFoundError(DKIMException):
 class UnknownKeyTypeError(DKIMException):
     """ Key type (k tag) is not known (rsa/ed25519) """
 
+class DnsTimeoutError(DKIMException):
+    """ DNS query for public key timed out """
+
 
 def select_headers(headers, include_headers):
     """Select message header fields to be signed/verified.
@@ -286,6 +298,16 @@ def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'
     if b'cv' in sig and sig[b'cv'] not in (CV_Pass, CV_Fail, CV_None):
         raise ValidationError("cv= value is not valid (%s)" % sig[b'cv'])
 
+    # Limit domain validation to ASCII domains because too hard
+    try:
+        str(sig[b'd'], 'ascii')
+        # No specials, which is close enough
+        if re.findall(b"[\(\)<>\[\]:;@\\,]", sig[b'd']):
+            raise ValidationError("d= value is not valid (%s)" % sig[b'd'])
+    except UnicodeDecodeError as e:
+        # Not an ASCII domain
+        pass
+
     # Nasty hack to support both str and bytes... check for both the
     # character and integer values.
     if not arc and b'i' in sig and (
@@ -446,8 +468,6 @@ def evaluate_pk(name, s):
               pk = nacl.signing.VerifyKey(pub[b'p'], encoder=nacl.encoding.Base64Encoder)
           except NameError:
               raise NaClNotFoundError('pynacl module required for ed25519 signing, see README.md')
-          except nacl.exceptions.ValueError as e:
-              raise KeyFormatError("could not parse ed25519 public key (%s): %s" % (pub[b'p'],e))
           keysize = 256
           ktag = b'ed25519'
   except KeyError:
@@ -457,9 +477,9 @@ def evaluate_pk(name, s):
           pk = parse_public_key(base64.b64decode(pub[b'p']))
           keysize = bitsize(pk['modulus'])
       except KeyError:
-          raise KeyFormatError("incomplete RSA public key: %s" % s)
+          raise KeyFormatError("incomplete public key: %s" % s)
       except (TypeError,UnparsableKeyError) as e:
-          raise KeyFormatError("could not parse RSA public key (%s): %s" % (pub[b'p'],e))
+          raise KeyFormatError("could not parse public key (%s): %s" % (pub[b'p'],e))
       ktag = b'rsa'
   if pub[b'k'] != b'rsa' and pub[b'k'] != b'ed25519':
       raise KeyFormatError('unknown algorithm in k= tag: {0}'.format(pub[b'k']))
@@ -780,6 +800,10 @@ class DomainSigner(object):
     except binascii.Error as e:
       self.logger.error('KeyFormatError: {0}'.format(e))
       return False
+    except dns.exception.Timeout as e:
+      self.logger.error('DnsTimeoutError: Domain: {0} Selector: {1} Error message: {2}'.format(
+          sig[b'd'], sig[b's'], e))
+      return False
     return self.verify_sig_process(sig, include_headers, sig_header, dnsfunc)
 
 
@@ -1013,10 +1037,10 @@ class ARC(DomainSigner):
     self.add_should_not(('Authentication-Results',))
     # check if authres has been imported
     try:
-        AuthenticationResultsHeader
+        authres.AuthenticationResultsHeader
     except:
         self.logger.debug("authres package not installed")
-        raise AuthresNotFoundError
+        raise authres.AuthresNotFoundError
 
     try:
         pk = parse_pem_private_key(privkey)
@@ -1025,8 +1049,14 @@ class ARC(DomainSigner):
 
     # extract, parse, filter & group AR headers
     ar_headers = [res.strip() for [ar, res] in self.headers if ar == b'Authentication-Results']
-    grouped_headers = [(res, AuthenticationResultsHeader.parse('Authentication-Results: ' + res.decode('utf-8')))
-                       for res in ar_headers]
+
+    grouped_headers = []
+    for res in ar_headers:
+        try: # see LP: #1884044
+            grouped_headers.append((res, authres.AuthenticationResultsHeader.parse('Authentication-Results: ' + res.decode('utf-8'))))
+        except authres.core.SyntaxError:
+            # Skip over invalid AR header fields
+            pass
     auth_headers = [res for res in grouped_headers if res[1].authserv_id == srv_id.decode('utf-8')]
 
     if len(auth_headers) == 0:
@@ -1040,7 +1070,7 @@ class ARC(DomainSigner):
     auth_results = srv_id + b'; ' + (b';' + self.linesep + b'  ').join(results)
 
     # extract cv
-    parsed_auth_results = AuthenticationResultsHeader.parse('Authentication-Results: ' + auth_results.decode('utf-8'))
+    parsed_auth_results = authres.AuthenticationResultsHeader.parse('Authentication-Results: ' + auth_results.decode('utf-8'))
     arc_results = [res for res in parsed_auth_results.results if res.method == 'arc']
     if len(arc_results) == 0:
       chain_validation_status = CV_None
@@ -1363,7 +1393,7 @@ def verify(message, logger=None, dnsfunc=get_txt, minkey=1024,
 
 
 # aiodns requires Python 3.5+, so no async before that
-if sys.version_info >= (3, 5):
+if sys.version_info >= (3, 5) and USE_ASYNC:
     try:
         import aiodns
         from dkim.asyncsupport import verify_async
